from __future__ import annotations

from digitalhub.entities.trigger._base.status import TriggerStatus


class TriggerStatusScheduler(TriggerStatus):
    """
    TriggerStatusScheduler status.
    """



from __future__ import annotations

import typing

from digitalhub.utils.exceptions import BuilderError

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.builder import EntityBuilder
    from digitalhub.entities._base.entity.entity import Entity
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities._base.entity.spec import Spec, SpecValidator
    from digitalhub.entities._base.entity.status import Status
    from digitalhub.entities._base.runtime_entity.builder import RuntimeEntityBuilder
    from digitalhub.runtimes._base import Runtime
    from digitalhub.runtimes.builder import RuntimeBuilder


class Factory:
    """
    Factory for creating and managing entity and runtime builders.

    This class implements the Factory pattern to manage the creation of
    entities and runtimes through their respective builders. It maintains
    separate registries for entity and runtime builders.

    Many function arguments are called kind_to_build_from to avoid overwriting
    kind in kwargs.

    Attributes
    ----------
    _entity_builders : dict[str, EntityBuilder | RuntimeEntityBuilder]
        Registry of entity builders indexed by kind.
    _runtime_builders : dict[str, RuntimeBuilder]
        Registry of runtime builders indexed by kind.

    Notes
    -----
    All builder methods may raise BuilderError if the requested kind
    is not found in the registry.
    """

    def __init__(self):
        self._entity_builders: dict[str, EntityBuilder | RuntimeEntityBuilder] = {}
        self._runtime_builders: dict[str, RuntimeBuilder] = {}

    def add_entity_builder(self, name: str, builder: EntityBuilder | RuntimeEntityBuilder) -> None:
        """
        Register an entity builder.

        Parameters
        ----------
        name : str
            The unique identifier for the builder.
        builder : EntityBuilder | RuntimeEntityBuilder
            The builder instance to register.

        Raises
        ------
        BuilderError
            If a builder with the same name already exists.
        """
        if name in self._entity_builders:
            raise BuilderError(f"Builder {name} already exists.")
        self._entity_builders[name] = builder()

    def add_runtime_builder(self, name: str, builder: RuntimeBuilder) -> None:
        """
        Register a runtime builder.

        Parameters
        ----------
        name : str
            The unique identifier for the builder.
        builder : RuntimeBuilder
            The builder instance to register.

        Raises
        ------
        BuilderError
            If a builder with the same name already exists.
        """
        if name in self._runtime_builders:
            raise BuilderError(f"Builder {name} already exists.")
        self._runtime_builders[name] = builder()

    def build_entity_from_params(self, **kwargs) -> Entity:
        """
        Build an entity from parameters.

        Parameters
        ----------
        **kwargs
            Entity parameters.

        Returns
        -------
        Entity
            Entity object.
        """
        try:
            kind = kwargs["kind"]
        except KeyError:
            raise BuilderError("Missing 'kind' parameter.")
        self._raise_if_entity_builder_not_found(kind)
        return self._entity_builders[kind].build(**kwargs)

    def build_entity_from_dict(self, obj: dict) -> Entity:
        """
        Build an entity from a dictionary.

        Parameters
        ----------
        obj : dict
            Dictionary with entity data.

        Returns
        -------
        Entity
            Entity object.
        """
        try:
            kind = obj["kind"]
        except KeyError:
            raise BuilderError("Missing 'kind' parameter.")
        self._raise_if_entity_builder_not_found(kind)
        return self._entity_builders[kind].from_dict(obj)

    def build_spec(self, kind_to_build_from: str, **kwargs) -> Spec:
        """
        Build an entity spec.

        Parameters
        ----------
        kind_to_build_from : str
            Entity type.

        Returns
        -------
        Spec
            Spec object.
        """
        self._raise_if_entity_builder_not_found(kind_to_build_from)
        return self._entity_builders[kind_to_build_from].build_spec(**kwargs)

    def build_metadata(self, kind_to_build_from: str, **kwargs) -> Metadata:
        """
        Build an entity metadata.

        Parameters
        ----------
        kind_to_build_from : str
            Entity type.

        Returns
        -------
        Metadata
            Metadata object.
        """
        self._raise_if_entity_builder_not_found(kind_to_build_from)
        return self._entity_builders[kind_to_build_from].build_metadata(**kwargs)

    def build_status(self, kind_to_build_from: str, **kwargs) -> Status:
        """
        Build an entity status.

        Parameters
        ----------
        kind_to_build_from : str
            Entity type.

        Returns
        -------
        Status
            Status object.
        """
        self._raise_if_entity_builder_not_found(kind_to_build_from)
        return self._entity_builders[kind_to_build_from].build_status(**kwargs)

    def build_runtime(self, kind_to_build_from: str, project: str) -> Runtime:
        """
        Build a runtime.

        Parameters
        ----------
        kind_to_build_from : str
            Runtime type.
        project : str
            Project name.

        Returns
        -------
        Runtime
            Runtime object.
        """
        self._raise_if_runtime_builder_not_found(kind_to_build_from)
        return self._runtime_builders[kind_to_build_from].build(project=project)

    def get_entity_type_from_kind(self, kind: str) -> str:
        """
        Get entity type from builder.

        Parameters
        ----------
        kind : str
            Entity type.

        Returns
        -------
        str
            Entity type.
        """
        self._raise_if_entity_builder_not_found(kind)
        return self._entity_builders[kind].get_entity_type()

    def get_executable_kind(self, kind: str) -> str:
        """
        Get executable kind.

        Parameters
        ----------
        kind : str
            Kind.

        Returns
        -------
        str
            Executable kind.
        """
        self._raise_if_entity_builder_not_found(kind)
        return self._entity_builders[kind].get_executable_kind()

    def get_action_from_task_kind(self, kind: str, task_kind: str) -> str:
        """
        Get action from task.

        Parameters
        ----------
        kind : str
            Kind.
        task_kind : str
            Task kind.

        Returns
        -------
        str
            Action.
        """
        self._raise_if_entity_builder_not_found(kind)
        return self._entity_builders[kind].get_action_from_task_kind(task_kind)

    def get_task_kind_from_action(self, kind: str, action: str) -> list[str]:
        """
        Get task kinds from action.

        Parameters
        ----------
        kind : str
            Kind.
        action : str
            Action.

        Returns
        -------
        list[str]
            Task kinds.
        """
        self._raise_if_entity_builder_not_found(kind)
        return self._entity_builders[kind].get_task_kind_from_action(action)

    def get_run_kind(self, kind: str) -> str:
        """
        Get run kind.

        Parameters
        ----------
        kind : str
            Kind.

        Returns
        -------
        str
            Run kind.
        """
        self._raise_if_entity_builder_not_found(kind)
        return self._entity_builders[kind].get_run_kind()

    def get_all_kinds(self, kind: str) -> list[str]:
        """
        Get all kinds.

        Parameters
        ----------
        kind : str
            Kind.

        Returns
        -------
        list[str]
            All kinds.
        """
        return self._entity_builders[kind].get_all_kinds()

    def get_spec_validator(self, kind: str) -> SpecValidator:
        """
        Get spec validators.

        Parameters
        ----------
        kind : str
            Kind.

        Returns
        -------
        SpecValidator
            Spec validator.
        """
        self._raise_if_entity_builder_not_found(kind)
        return self._entity_builders[kind].get_spec_validator()

    def _raise_if_entity_builder_not_found(self, kind: str) -> None:
        """
        Verify entity builder existence.

        Parameters
        ----------
        kind : str
            The entity kind to verify.

        Raises
        ------
        BuilderError
            If no builder exists for the specified kind.
        """
        if kind not in self._entity_builders:
            raise BuilderError(f"Entity builder for kind '{kind}' not found.")

    def _raise_if_runtime_builder_not_found(self, kind: str) -> None:
        """
        Verify runtime builder existence.

        Parameters
        ----------
        kind : str
            The runtime kind to verify.

        Raises
        ------
        BuilderError
            If no builder exists for the specified kind.
        """
        if kind not in self._runtime_builders:
            raise BuilderError(f"Runtime builder for kind '{kind}' not found.")


factory = Factory()



from __future__ import annotations

import typing
from concurrent.futures import ThreadPoolExecutor

from digitalhub.entities._base.executable.entity import ExecutableEntity
from digitalhub.entities._commons.enums import EntityTypes, Relationship
from digitalhub.factory.factory import factory
from digitalhub.utils.exceptions import BackendError

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities._base.entity.spec import SpecValidator
    from digitalhub.entities.function._base.spec import FunctionSpec
    from digitalhub.entities.function._base.status import FunctionStatus
    from digitalhub.entities.run._base.entity import Run
    from digitalhub.entities.trigger._base.entity import Trigger


class Function(ExecutableEntity):
    """
    A class representing a function.
    """

    ENTITY_TYPE = EntityTypes.FUNCTION.value

    def __init__(
        self,
        project: str,
        name: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: FunctionSpec,
        status: FunctionStatus,
        user: str | None = None,
    ) -> None:
        super().__init__(project, name, uuid, kind, metadata, spec, status, user)

        self.spec: FunctionSpec
        self.status: FunctionStatus

    ##############################
    #  Function Methods
    ##############################

    def run(
        self,
        action: str,
        local_execution: bool = False,
        wait: bool = False,
        log_info: bool = True,
        **kwargs,
    ) -> Run:
        """
        Run function. This method creates a new run and executes it.

        Parameters
        ----------
        action : str
            Action to execute.
        local_execution : bool
            Flag to determine if object has local execution.
        wait : bool
            Flag to wait for execution.
        log_info : bool
            Flag to log information while waiting.
        **kwargs : dict
            Keyword arguments passed to Run builder.

        Returns
        -------
        Run
            Run instance.
        """
        # Get task and run kind
        task_kind = factory.get_task_kind_from_action(self.kind, action)
        run_kind = factory.get_run_kind(self.kind)

        # Create or update new task
        task = self._get_or_create_task(task_kind)

        # Run function from task
        run = task.run(run_kind, save=False, local_execution=local_execution, **kwargs)

        # Set as run's parent
        run.add_relationship(Relationship.RUN_OF.value, self.key)
        run.save()

        # If execution is done by DHCore backend, return the object
        if not local_execution:
            if self._context().local:
                raise BackendError("Cannot run remote function with local backend.")
            if wait:
                return run.wait(log_info=log_info)
            return run

        # If local execution, build and launch run.
        # Detach the run from the main thread
        run.build()
        with ThreadPoolExecutor(max_workers=1) as executor:
            result = executor.submit(run.run)
            r = result.result()
        return r

    def trigger(self, action: str, trigger_kind: str, trigger_name: str, **kwargs) -> Trigger:
        """
        Trigger function.

        Parameters
        ----------
        action : str
            Action to execute.
        trigger_kind : str
            Trigger kind.
        **kwargs : dict
            Keyword arguments passed to Run builder.

        Returns
        -------
        Run
            Run instance.
        """
        # Get task
        task_kind = factory.get_task_kind_from_action(self.kind, action)
        task = self._get_or_create_task(task_kind)
        task_string = task._get_task_string()

        # Get run validator for building trigger template
        run_kind = factory.get_run_kind(self.kind)
        run_validator: SpecValidator = factory.get_spec_validator(run_kind)
        if kwargs is None:
            kwargs = {}
        kwargs[self.ENTITY_TYPE] = self._get_executable_string()
        kwargs["task"] = task_string

        template = run_validator(**kwargs).to_dict()

        # Override kwargs
        kwargs["project"] = self.project
        kwargs["kind"] = trigger_kind
        kwargs["name"] = trigger_name
        kwargs["template"] = template

        # Create object instance
        trigger: Trigger = factory.build_entity_from_params(**kwargs)
        trigger.save()
        return trigger



from __future__ import annotations

from enum import Enum


class CredsOrigin(Enum):
    """
    List credential origins.
    """

    ENV = "env"
    FILE = "file"


class SetCreds(Enum):
    """
    List supported environments.
    """

    DEFAULT = "__default"
    DH_ENV = "DH_NAME"



from __future__ import annotations

from digitalhub.entities._base.entity.spec import Spec, SpecValidator


class TriggerSpec(Spec):
    """
    TriggerSpec specifications.
    """

    def __init__(self, task: str, function: str, template: dict) -> None:
        super().__init__()
        self.task = task
        self.function = function
        self.template = template


class TriggerValidator(SpecValidator):
    """
    TriggerValidator validator.
    """

    task: str
    """Task string."""

    function: str
    """Function string."""

    template: dict
    """Template string."""



from __future__ import annotations

from abc import abstractmethod

from digitalhub.entities._commons.enums import ApiCategories


class ClientKeyBuilder:
    """
    Class that build the key of entities.
    """

    def build_key(self, category: str, *args, **kwargs) -> str:
        """
        Build key.

        Parameters
        ----------
        category : str
            Key category.
        *args : tuple
            Positional arguments.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        str
            Key.
        """
        if category == ApiCategories.BASE.value:
            return self.base_entity_key(*args, **kwargs)
        return self.context_entity_key(*args, **kwargs)

    @abstractmethod
    def base_entity_key(self, entity_id: str) -> str:
        """
        Build for base entity key.
        """

    @abstractmethod
    def context_entity_key(
        self,
        project: str,
        entity_type: str,
        entity_kind: str,
        entity_name: str,
        entity_id: str | None = None,
    ) -> str:
        """
        Build for context entity key.
        """



from __future__ import annotations

from digitalhub.context.api import get_context
from digitalhub.stores.configurator.configurator import configurator
from digitalhub.stores.data.enums import StoreEnv


def get_default_store(project: str) -> str:
    """
    Get default store URI.

    Parameters
    ----------
    project : str
        Project name.

    Returns
    -------
    str
        Default store URI.
    """
    context = get_context(project)
    store = context.config.get(StoreEnv.DEFAULT_FILES_STORE.value.lower())
    if store is not None:
        return store
    store = configurator.load_var(StoreEnv.DEFAULT_FILES_STORE.value)
    if store is None or store == "":
        raise ValueError(
            "No default store found. "
            "Please set a default store "
            f"in your environment (e.g. export {StoreEnv.DEFAULT_FILES_STORE.value}=) "
            "or set it in project config."
        )
    return store



from __future__ import annotations

import importlib
import pkgutil
import re
from types import ModuleType

from digitalhub.factory.factory import factory


def import_module(package: str) -> ModuleType:
    """
    Import a module dynamically by package name.

    Parameters
    ----------
    package : str
        The fully qualified package name to import.

    Returns
    -------
    ModuleType
        The imported module object.

    Raises
    ------
    ModuleNotFoundError
        If the specified package cannot be found.
    RuntimeError
        If any other error occurs during import.
    """
    try:
        return importlib.import_module(package)
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(f"Package {package} not found.") from e
    except Exception as e:
        raise RuntimeError(f"An error occurred while importing {package}.") from e


def list_runtimes() -> list[str]:
    """
    List all installed DigitalHub runtime packages.

    Scans installed packages for those matching the pattern
    'digitalhub_runtime_*'.

    Returns
    -------
    list[str]
        List of runtime package names.

    Raises
    ------
    RuntimeError
        If an error occurs while scanning for runtime packages.
    """
    pattern = r"digitalhub_runtime_.*"
    runtimes: list[str] = []
    try:
        for _, name, _ in pkgutil.iter_modules():
            if re.match(pattern, name):
                runtimes.append(name)
        return runtimes
    except Exception as e:
        raise RuntimeError("Error listing installed runtimes.") from e


def register_runtimes_entities() -> None:
    """
    Register all runtime builders and their entities into the factory.

    Imports each runtime package and registers its entity and runtime
    builders with the global factory instance.
    """
    for package in list_runtimes():
        module = import_module(package)
        entity_builders = getattr(module, "entity_builders")
        for entity_builder_tuple in entity_builders:
            kind, builder = entity_builder_tuple
            factory.add_entity_builder(kind, builder)

        runtime_builders = getattr(module, "runtime_builders")
        for runtime_builder_tuple in runtime_builders:
            kind, builder = runtime_builder_tuple
            factory.add_runtime_builder(kind, builder)


def register_entities() -> None:
    """
    Register core entity builders into the factory.

    Imports the core entities module and registers all entity builders
    with the global factory instance.

    Raises
    ------
    RuntimeError
        If registration of core entities fails.
    """
    try:
        module = import_module("digitalhub.entities.builders")
        entities_builders_list = getattr(module, "entity_builders")
        for entity_builder_tuple in entities_builders_list:
            kind, builder = entity_builder_tuple
            factory.add_entity_builder(kind, builder)
    except Exception as e:
        raise RuntimeError("Error registering entities.") from e



from __future__ import annotations

import typing
from pathlib import Path

if typing.TYPE_CHECKING:
    from digitalhub.entities.project._base.entity import Project
    from digitalhub.stores.client._base.client import Client


class Context:
    """
    Context class built from a Project instance.

    Contains project-specific information and state, including project name,
    client instance, local context paths, and run-time information.

    Attributes
    ----------
    name : str
        The name of the project.
    client : BaseClient
        The client instance (local or remote) associated with the project.
    config : dict
        Project configuration profile.
    local : bool
        Whether the client is local or remote.
    root : Path
        The local context project path.
    is_running : bool
        Flag indicating if the context has an active run.
    _run_ctx : str | None
        Current run key, if any.
    """

    def __init__(self, project: Project) -> None:
        self.name: str = project.name
        self.client: Client = project._client
        self.config: dict = project.spec.config
        self.local: bool = project._client.is_local()
        self.root: Path = Path(project.spec.context)
        self.root.mkdir(parents=True, exist_ok=True)

        self.is_running: bool = False
        self._run_ctx: str | None = None

    def set_run(self, run_ctx: str) -> None:
        """
        Set the current run key.

        Parameters
        ----------
        run_ctx : str
            The run key to set.
        """
        self.is_running = True
        self._run_ctx = run_ctx

    def unset_run(self) -> None:
        """
        Clear the current run key and reset running state.
        """
        self.is_running = False
        self._run_ctx = None

    def get_run_ctx(self) -> str | None:
        """
        Get the current run key.

        Returns
        -------
        str | None
            The current run key if set, None otherwise.
        """
        return self._run_ctx



from __future__ import annotations

from pathlib import Path
from urllib.parse import urlparse

from boto3 import client as boto3_client

from digitalhub.stores.configurator.enums import CredsOrigin
from digitalhub.stores.data.s3.configurator import S3StoreConfigurator
from digitalhub.utils.exceptions import StoreError


def get_bucket_name(path: str) -> str:
    """
    Get bucket name from path.

    Parameters
    ----------
    path : str
        The source path to get the key from.

    Returns
    -------
    str
        The bucket name.
    """
    return urlparse(path).netloc


def get_bucket_and_key(path: str) -> tuple[str, str]:
    """
    Get bucket and key from path.

    Parameters
    ----------
    path : str
        The source path to get the key from.

    Returns
    -------
    tuple[str, str]
        The bucket and key.
    """
    parsed = urlparse(path)
    return parsed.netloc, parsed.path


def get_s3_source(bucket: str, key: str, filename: Path) -> None:
    """
    Get S3 source.

    Parameters
    ----------
    bucket : str
        S3 bucket name.
    key : str
        S3 object key.
    filename : Path
        Path where to save the function source.

    Returns
    -------
    None
    """
    # Try to get client from environment variables
    try:
        cfg = S3StoreConfigurator().get_boto3_client_config(CredsOrigin.ENV.value)
        s3 = boto3_client("s3", **cfg)
        s3.download_file(bucket, key, filename)

    # Fallback to file
    except StoreError:
        cfg = S3StoreConfigurator().get_boto3_client_config(CredsOrigin.FILE.value)
        s3.download_file(bucket, key, filename)



from __future__ import annotations

from digitalhub.entities.trigger._base.spec import TriggerSpec, TriggerValidator


class TriggerSpecScheduler(TriggerSpec):
    """
    TriggerSpecScheduler specifications.
    """

    def __init__(self, task: str, function: str, template: dict, schedule: str) -> None:
        super().__init__(task, function, template)
        self.schedule = schedule


class TriggerValidatorScheduler(TriggerValidator):
    """
    TriggerValidatorScheduler validator.
    """

    schedule: str
    """Quartz cron expression."""
